/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993-1994 G. Wiesenekker
	E-mail: wiesenekker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "zzzzzz.h"
local char *FILE_NAME = __FILE__;

#ifndef DEBUG_LEVEL
	#define DEBUG_LEVEL 1
#endif

void pgn_version(char *version_string)
{
	sprintf(version_string, "%s %s %d", __DATE__, __TIME__, DEBUG_LEVEL);
}

/*
b-tree stuff
*/

/*
The order of the btree is the maximum number of pointers in a node.
The maximum number of entries is BTREE_ORDER - 1.
To all entries but the last there is associated a pointer to a btree with
entries with values LESS THAN entry.
To the last entry there is associated a pointer to a btree with values
GREATER THAN the last entry.
For symmetry, we include a special entry[BTREE_ORDER] with value LONG_MAX,
so now to all entries there is associated a pointer to a btree with values 
LESS THAN that entry. We have to split a node if entry[BTREE_ORDER] != LONG_MAX.
We can not store an entry with value LONG_MAX.
*/

#define BTREE_ORDER 513
#if ((BTREE_ORDER % 2) == 0)
#error BTREE_ORDER should be odd!
#endif
#if (BTREE_ORDER < 3)
#error BTREE_ORDER should be greater than three!
#endif

/*
BTREE_DEPTH_MAX is the maximum number of nodes that must be searched
to find an entry. As the order of a btree is at least 3, 32 allows
for 3^32 entries.
*/

#define BTREE_DEPTH_MAX 32

/*
btree_entry are the 32 bit CRC checksums of the FEN positions.
dbase_offset are pointers to information about the positions.
btree_child are pointers to child nodes.
*/

typedef struct btree_node_tag
{
	long btree_entry[BTREE_ORDER + 1];
	long dbase_offset[BTREE_ORDER + 1];
	long btree_child[BTREE_ORDER + 1];
} btree_node_t;

/*
MAGIC and VERSION are used to identify the files.
*/

#define PGN_MAGIC     1
#define PGN_VERSION   1

#define DBASE_MAGIC   2
#define DBASE_VERSION 1

local FILE* fkey;
local FILE* fbin;

/*
btree_node contains the current node.
btree_node_offset points to the current node.
*/
 
local btree_node_t btree_node;
local long btree_node_offset;

local long nodes, nstored;

local void display_btree(int, long);
local void display_btree(int n, long btree_temp_offset)
{
	int i, j;
	btree_node_t btree_temp;
	
	if (btree_temp_offset == LONG_MAX) return;
	FSEEK(fkey, btree_temp_offset)
	FREAD(fkey, &btree_temp, btree_node_t, 1)
	for (i = 0; i < BTREE_ORDER + 1; i++)
	{
		for (j = 0; j < n; j++)
			my_printf(" ");
		my_printf("(%d)%ld\n", n, btree_temp.btree_entry[i]);
		display_btree(n + 1, btree_temp.btree_child[i]);
	}
}

local void check_btree(int, long );
local void check_btree(int n, long btree_temp_offset)
{
	int i;
	btree_node_t btree_temp;
	
	if (btree_temp_offset == LONG_MAX) return;
	nodes++;
	FSEEK(fkey, btree_temp_offset)
	FREAD(fkey, &btree_temp, btree_node_t, 1)
	for (i = 0; i < BTREE_ORDER - 1; i++)
	{
		BUG(btree_temp.btree_entry[i] > btree_temp.btree_entry[i + 1])
		if (btree_temp.btree_entry[i] != LONG_MAX) nstored++;
	}
	BUG(btree_temp.btree_entry[BTREE_ORDER - 1] != LONG_MAX)
	BUG(btree_temp.btree_entry[BTREE_ORDER] != LONG_MAX)
	for (i = 0; i < BTREE_ORDER; i++)
		check_btree(n + 1, btree_temp.btree_child[i]);
}

/*
open_dbase opens the dbase and checks the version and
updates the dbase if it is an old version.
*/

local void open_dbase(char *dbase_name, uchar_t, uchar_t);
local void open_dbase(char *dbase_name, uchar_t magic, uchar_t version)
{
	long i;
	char key_name[FILENAME_MAX];
	char bin_name[FILENAME_MAX];
	uchar_t temp_magic, temp_version;
	
	strcpy(key_name, dbase_name);
	strcat(key_name, ".key");
	strcpy(bin_name, dbase_name);
	strcat(bin_name, ".bin");
	my_printf("%s %s\n", key_name, bin_name);
	
	if ((fkey = fopen(key_name, "r+b")) == NULL)
	{
		BUG((fkey = fopen(key_name, "wb")) == NULL)
		btree_node_offset = LONG_MAX;
		FWRITE(fkey, &(btree_node_offset), long, 1)
		FTELL(fkey, btree_node_offset)
		FBOF(fkey)
		FWRITE(fkey, &btree_node_offset, long, 1)

		nodes = 1;
		nstored = 0;
				
		/*
		Clear the node.
		*/
		
		for (i = 0; i < BTREE_ORDER + 1; i++)
		{
			btree_node.btree_entry[i] = LONG_MAX;
			btree_node.dbase_offset[i] = LONG_MAX;
			btree_node.btree_child[i] = LONG_MAX;
		}
		FWRITE(fkey, &btree_node, btree_node_t, 1)
		FCLOSE(fkey)
		BUG((fbin = fopen(bin_name, "wb")) == NULL)
		FWRITE(fbin, &(magic), uchar_t, 1)
		FWRITE(fbin, &(version), uchar_t, 1)
		FCLOSE(fbin)
		BUG((fkey = fopen(key_name, "r+b")) == NULL)
	}
	FREAD(fkey, &btree_node_offset, long, 1)
	nodes = nstored = 0;
	check_btree(0, btree_node_offset);
	my_printf("nodes=%ld stored=%ld\n", nodes, nstored);
	BUG((fbin = fopen(bin_name, "r+b")) == NULL)
	FREAD(fbin, &temp_magic, uchar_t, 1)
	FREAD(fbin, &temp_version, uchar_t, 1)
	if (temp_version != version)
		FATAL_ERROR("No converters available, complain to G. Wiesenekker")
	my_printf("offset=%ld magic=%u version=%u\n",
		btree_node_offset, (uint_t) magic, (uint_t) version);
}

local void close_dbase(void);
local void close_dbase(void)
{
	FCLOSE(fkey)
	FCLOSE(fbin)
}

/*
binary_search returns an index i such that 
value <= entry[i]
*/

local long binary_search(long);
local long binary_search(long value)
{
	long jlow, jhigh, jmiddle;
	static int ok = ((13 / 2) == (13 >> 1));
	
	jlow = 0;
	jhigh = BTREE_ORDER - 1;
	
	if (ok)
	{
		while((jmiddle = (jlow + jhigh) >> 1) != jlow)
		{
			if (value > btree_node.btree_entry[jmiddle])
				jlow = jmiddle;
			else
				jhigh = jmiddle;
		}
	}
	else
	{
		while((jmiddle = (jlow + jhigh) / 2) != jlow)
		{
			if (value > btree_node.btree_entry[jmiddle])
				jlow = jmiddle;
			else
				jhigh = jmiddle;
		}
	}
	BUG(jlow > jhigh)
	if (value <= btree_node.btree_entry[jlow])
		return(jlow);
	else if (value <= btree_node.btree_entry[jhigh])
		return(jhigh);
	else
		return(INVALID);
}

/*
btree_search searches for a value in the btree.
If the value is found, btree_node is the node that contains the value,
and btree_index is the entry.
If the value is not found, btree_node is the node in which the value is 
to be inserted, and btree_index is the entry BEFORE which the value is
to be inserted.

btree_chain contains the pointers to the nodes that have been 
searched. It is needed for inserting a value into the btree.
*/

local long btree_chain[BTREE_DEPTH_MAX];
local int btree_depth, btree_index;

local int btree_search(long);
local int btree_search(long value)
{
	int i;
	long btree_next;
	
	/*
	values equal to LONG_MAX are not allowed.
	*/
	
	BUG(value == LONG_MAX)
	
	/*
	We start at the root.
	*/
	
	FBOF(fkey)
	FREAD(fkey, &btree_node_offset, long, 1)
	FSEEK(fkey, btree_node_offset)
	FREAD(fkey, &btree_node, btree_node_t, 1)
	btree_next = LONG_MAX;
	btree_depth = 0;
	btree_chain[btree_depth] = LONG_MAX;
	
	while(TRUE)
	{
		/*
		Search the current node
		*/

		i = binary_search(value);
		
		if (value == btree_node.btree_entry[i])
		{
			btree_index = i;
			return(TRUE);
		}
		
		BUG(value > btree_node.btree_entry[i])
		btree_next = btree_node.btree_child[i];
		
		/*
		A leaf is there is no child.
		*/
		
		if (btree_next == LONG_MAX)
		{
			btree_index = i;
			return(FALSE);
		}
		else
		{
			BUG(btree_depth == BTREE_DEPTH_MAX)
			btree_chain[btree_depth++] = btree_node_offset;
			btree_node_offset = btree_next;
			FSEEK(fkey, btree_node_offset)
			FREAD(fkey, &btree_node, btree_node_t, 1)
		}
	}
}

/*
btree_insert inserts a value into the btree.
*/

local void btree_insert(long, long);
local void btree_insert(long value, long where)
{
	int i, m;
	btree_node_t btree_new;
	long btree_new_offset;
	
	btree_new_offset = LONG_MAX;
	while(TRUE)
	{
		/*
		btree_index is the entry BEFORE which the value is to be inserted.
		We make room for the entry.
		*/
		
		for (i = BTREE_ORDER; i > btree_index; i--)
		{
			btree_node.btree_entry[i] = btree_node.btree_entry[i - 1];
			btree_node.dbase_offset[i] = btree_node.dbase_offset[i - 1];
			btree_node.btree_child[i] = btree_node.btree_child[i - 1];
		}
		
		btree_node.btree_entry[btree_index] = value;
		btree_node.dbase_offset[btree_index] = where;
		btree_node.btree_child[btree_index] = btree_new_offset;
	
		if (btree_node.btree_entry[BTREE_ORDER - 1] == LONG_MAX)
		{
			FSEEK(fkey, btree_node_offset)
			FWRITE(fkey, &btree_node, btree_node_t, 1)
			break;
		}
		
		/*
		We split the node if the node is full.
		We make a copy of the current node.
		*/
		
		btree_new = btree_node;
		FEOF(fkey)
		FTELL(fkey, btree_new_offset)
		
		/*
		The new node will contain all values LESS THAN the middle value.
		The old node will contain all values GREATER THAN the middle value.
		*/
		
		m = BTREE_ORDER / 2;
		value = btree_node.btree_entry[m];
		where = btree_node.dbase_offset[m];
	
		/*
		We zap all entries greater than the middle value in the new node.
		*/
		
		btree_new.btree_entry[m] = LONG_MAX;
		btree_new.dbase_offset[m] = LONG_MAX;
		for (i = m + 1; i < BTREE_ORDER + 1; i++)
		{
			btree_new.btree_entry[i] = LONG_MAX;
			btree_new.dbase_offset[i] = LONG_MAX;
			btree_new.btree_child[i] = LONG_MAX;
		}
		
		/*
		And we zap all entries less than the middle value in the current node.
		*/
		
		/*
		Move the values.
		*/
		
		for (i = m + 1; i < BTREE_ORDER + 1; i++)
		{
			btree_node.btree_entry[i - m - 1] = btree_node.btree_entry[i];
			btree_node.dbase_offset[i - m - 1] = btree_node.dbase_offset[i];
			btree_node.btree_child[i - m - 1] = btree_node.btree_child[i];
		}
		
		for (i = m + 1; i < BTREE_ORDER + 1; i++)
		{
			btree_node.btree_entry[i] = LONG_MAX;
			btree_node.dbase_offset[i] = LONG_MAX;
			btree_node.btree_child[i] = LONG_MAX;
		}
		FWRITE(fkey, &btree_new, btree_node_t, 1)
		
		FSEEK(fkey, btree_node_offset)
		FWRITE(fkey, &btree_node, btree_node_t, 1)
				
		if (btree_depth == 0)
		{
			/*
			Create a new root node
			*/
			
			for (i = 0; i < BTREE_ORDER + 1; i++)
			{
				btree_node.btree_entry[i] = LONG_MAX;
				btree_node.dbase_offset[i] = LONG_MAX;
				btree_node.btree_child[i] = LONG_MAX;
			}
			btree_node.btree_child[0] = btree_node_offset;
			FEOF(fkey)
			FTELL(fkey, btree_node_offset)
			FWRITE(fkey, &btree_node, btree_node_t, 1)
			FBOF(fkey)
			FWRITE(fkey, &btree_node_offset, long, 1)
			btree_index = 0;
		}
		else
		{
			btree_node_offset = btree_chain[--btree_depth];
			BUG(btree_node_offset == LONG_MAX)
			FSEEK(fkey, btree_node_offset)
			FREAD(fkey, &btree_node, btree_node_t, 1)
			btree_index = binary_search(value);
		}
	}
}

local void btree_linear(long);
local void btree_linear(long n)
{
	long i, j;
	
	my_printf("linear fill\n");
	remove("btree.key");
	remove("btree.bin");
	open_dbase("btree", PGN_MAGIC, PGN_VERSION);
	for (j = 0; j < n; j++)
	{
		my_printf("storing %ld\n", j);
		if (btree_search(j))
		{
			my_printf("found %ld!\n", j);
		}
		else
		{
			btree_insert(j, LONG_MAX);
			for (i = 0; i <= j; i++)
				BUG(!btree_search(i))
		}
	}
	FBOF(fkey)
	FREAD(fkey, &btree_node_offset, long, 1)
	/* display_btree(0, btree_node_offset); */
	nodes = nstored = 0;
	check_btree(0, btree_node_offset);
	my_printf("nodes=%ld stored=%ld\n", nodes, nstored);
	BUG(nstored != n)
	close_dbase();
}

local void btree_random(long);
local void btree_random(long n)
{
	long r;
	long i;
	
	my_printf("random fill\n");
	remove("btree.key");
	remove("btree.bin");
	open_dbase("btree", PGN_MAGIC, PGN_VERSION);
	for (i = 0; i < n; i++)
	{
		r = ulrand() % LONG_MAX;
		if (btree_search(r))
		{
			my_printf("found %ld!\n", r);
		}
		else
		{
			btree_insert(r, LONG_MAX);
			BUG(!btree_search(r))
		}
	}
	FBOF(fkey)
	FREAD(fkey, &btree_node_offset, long, 1)
	/* display_btree(0, btree_node_offset); */
	nodes = nstored = 0;
	check_btree(0, btree_node_offset);
	my_printf("nodes=%ld stored=%ld\n", nodes, nstored);
	BUG(nstored != n)
	close_dbase();
}

void btree(void)
{
	btree_linear(100);
	btree_random(1000);
}

/*
PGN stuff
*/

local long ngame;

#define PGN_LINE_MAX 256
local char pgn_value[PGN_LINE_MAX];
local int pgn_kind;
local long pgn_line;

#define PGN_EOF      1
#define PGN_STRING   2
#define PGN_INTEGER  3
#define PGN_PERIOD   4
#define PGN_RESULT   5
#define PGN_ASTERISK 6
#define PGN_LBRACKET 7
#define PGN_RBRACKET 8
#define PGN_LPAREN   9
#define PGN_RPAREN   10
#define PGN_NAG      11
#define PGN_OTHER    12

local char buf[LINE_MAX];
local char *pbuf;
local int in_comment;

local void get_next_pgn_token(FILE *);
local void get_next_pgn_token(FILE *f)
{
	char *p;
	
	label_next_char:
	if (*pbuf == '\0')
	{
		if (fgets(buf, PGN_LINE_MAX, f) == NULL)
		{
			pgn_kind = PGN_EOF;
			return;
		}
		pbuf = buf;
		pgn_line++;
	}
	
	/*
	space (or tab).
	*/
	
	while ((*pbuf == ' ') or (*pbuf == '\t')) pbuf++;
	
	/*
	end of line.
	*/
	
	if (*pbuf == '\0') goto label_next_char;
	
	/*
	new line.
	*/
	
	if (*pbuf == '\n')
	{
		pbuf++;
		goto label_next_char;
	}
	
	/*
	comment.
	*/
	
	if (in_comment)
	{
		if (*pbuf == '}') in_comment = FALSE;
		pbuf++;
		goto label_next_char;
	}

	/*
	comment.
	*/
	
	if (*pbuf == '{')
	{
		in_comment = TRUE;
		pbuf++;
		goto label_next_char;
	}
	
	/*
	end of line comment.
	*/
	
	if (*pbuf == ';') 
	{
		*pbuf = '\0';
		goto label_next_char;
	}
	
	/*
	escape.
	*/
	
	if (*pbuf == '%') 
	{
		*pbuf = '\0';
		goto label_next_char;
	}
	
	/*
	string.
	*/
	
	if (*pbuf == '\"')
	{
		p = pgn_value;
		pbuf++;
		while(TRUE)
		{
			if (*pbuf == '\0')
			{
				my_printf("unexpected end of string, line=%ld!\n", pgn_line);
				pgn_kind = PGN_EOF;
				return;
			}
			if (*pbuf == '\"')
			{
				*p = '\0';
				pgn_kind = PGN_STRING;
				pbuf++;
				return;
			}
			if (*pbuf == '\\') pbuf++;
			*p++ = *pbuf++;
		}
	}
	
	/*
	result.
	*/
	
	if (strncmp(pbuf, "1-0", 3) == 0)
	{
		strcpy(pgn_value, "1-0");
		pgn_kind = PGN_RESULT;
		pbuf += 3;
		return;
	}
	if (strncmp(pbuf, "1/2-1/2", 7) == 0)
	{
		strcpy(pgn_value, "1/2-1/2");
		pgn_kind = PGN_RESULT;
		pbuf += 7;
		return;
	}
	if (strncmp(pbuf, "0-1", 3) == 0)
	{
		strcpy(pgn_value, "0-1");
		pgn_kind = PGN_RESULT;
		pbuf += 3;
		return;
	}
	if (strncmp(pbuf, "*", 1) == 0)
	{
		strcpy(pgn_value, "*");
		pgn_kind = PGN_RESULT;
		pbuf += 1;
		return;
	}
	
	/*
	digit.
	*/
	
	if (isdigit(*pbuf))
	{
		p = pgn_value;
		while(TRUE)
		{
			*p++ = *pbuf++;
			if (!isdigit(*pbuf))
			{
				*p = '\0';
				pgn_kind = PGN_INTEGER;
				return;
			}
		}
	}
	
	/*
	period.
	*/
	
	if (*pbuf == '.')
	{
		pbuf++;
		pgn_kind = PGN_PERIOD;
		return;
	}
	
	/*
	asterisk.
	*/
	
	if (*pbuf == '*')
	{
		pbuf++;
		pgn_kind = PGN_ASTERISK;
		return;
	}
	
	/*
	bracket.
	*/
	
	if (*pbuf == '[')
	{	
		pbuf++;
		pgn_kind = PGN_LBRACKET;
		return;
	}
	if (*pbuf == ']')
	{
		pbuf++;
		pgn_kind = PGN_RBRACKET;
		return;
	}
	
	/*
	parenthesis.
	*/
	
	if (*pbuf == '(')
	{
		pbuf++;
		pgn_kind = PGN_LPAREN;
		return;
	}
	if (*pbuf == ')')
	{
		pbuf++;
		pgn_kind = PGN_RPAREN;
		return;
	}
	
	/*
	NAG.
	*/
	
	if (*pbuf == '$')
	{
		p = pgn_value;
		while(TRUE)
		{
			*p++ = *pbuf++;
			if (!isdigit(*pbuf))
			{
				*p = '\0';
				pgn_kind = PGN_NAG;
				return;
			}
		}
	}
	
	/*
	other.
	*/
	
	p = pgn_value;
	while(TRUE)
	{
		*p++ = *pbuf++;
		if (*pbuf == '\0')
		{
			pgn_kind = PGN_OTHER;
			return;
		}
		if ((*pbuf == ' ') or (*pbuf == '\n'))
		{
			*p = '\0';
			pgn_kind = PGN_OTHER;
			return;
		}
	}
}

local void debug_pgn(FILE *);
local void debug_pgn(FILE *f)
{
	while(TRUE)
	{
		get_next_pgn_token(f);
		if (pgn_kind == PGN_EOF) break;
		if (pgn_kind == PGN_STRING)
		{
			my_printf("string %s\n", pgn_value);
		}
		else if (pgn_kind == PGN_RESULT)
		{
			my_printf("result %s\n", pgn_value);
		}
		else if (pgn_kind == PGN_INTEGER)
		{
			my_printf("integer %s\n", pgn_value);
		}
		else if (pgn_kind == PGN_PERIOD)
		{
			my_printf(".\n");
		}
		else if (pgn_kind == PGN_LBRACKET)
		{
			my_printf("[\n");
		}
		else if (pgn_kind == PGN_RBRACKET)
		{
			my_printf("]\n");
		}
		else if (pgn_kind == PGN_LPAREN)
		{
			my_printf("(\n");
		}
		else if (pgn_kind == PGN_RPAREN)
		{
			my_printf(")\n");
		}
		else if (pgn_kind == PGN_NAG)
		{
			my_printf("nag\n");
		}
		else if (pgn_kind == PGN_OTHER)
		{
			my_printf("other %s\n", pgn_value);
		}
		else
		{
			FATAL_ERROR("pgn_kind")
		}
	}
}

#define SPACE  0x20
#define WIGGLE 0x7E
#define BASE   (WIGGLE - SPACE)

local void bin_to_asc(int, int *, char *);
local void bin_to_asc(int n, int *x, char *s)
{
	int i, k;
	
	for (i = 0; i < n; i++)
	{
		k = x[i];
		if (k < BASE)
			*s++ = k + SPACE;
		else
		{
			*s++ = WIGGLE;
			*s++ = k % BASE + SPACE;
			*s++ = k / BASE + SPACE;
		}
	}
	*s = '\0';
}

local void asc_to_bin(int *, int *, char *);
local void asc_to_bin(int *n, int *x, char *s)
{
	int i, j, k;
	
	i = 0;
	while((k = *s++) >= SPACE)
	{
		if (k < WIGGLE)
		{
			x[i++] = k - SPACE;
		}
		else
		{
			BUG(k != WIGGLE)
			k = *s++ - SPACE;
			j = *s++ - SPACE;
			BUG(k < 0)
			BUG(j < 0)
			x[i++] = k + j * BASE;
		}
	}
	*n = i;
}

local void do_some_moves(int, int *);
local void do_some_moves(int n, int *x)
{
	int i, imove, nmoves;
	move_t *base;
	
	for (i = 0; i < n; i++)
	{
		imove = x[i];
		base = capt_top;
		nmoves = gen_all_moves();
		BUG(imove >= nmoves)
		do_move(base[imove]);
		capt_top = base;
	}
}

local void undo_some_moves(int);
local void undo_some_moves(int n)
{
	int i;
	
	for (i = n; i > 0; i--)
		undo_move();
}
			
#define NGAME_MOVES_MAX 1024

local void comp_pgn(FILE *);
local void comp_pgn(FILE *f)
{
	int imove;
	long i;
	int game_moves[NGAME_MOVES_MAX];
	char game_string[NGAME_MOVES_MAX];
	char white_player[PGN_LINE_MAX];
	char black_player[PGN_LINE_MAX];
	char result[PGN_LINE_MAX];
	int ngame_moves;
	move_t *base;
	long r;
	
	open_dbase("pgn", PGN_MAGIC, PGN_VERSION);
	get_next_pgn_token(f);
	while(TRUE)
	{
		/*
		tags.
		*/
		
		i = 0;
		if (pgn_kind == PGN_EOF) break;
		while(TRUE)
		{
			if (pgn_kind != PGN_LBRACKET) break;
			
			/*
			symbol.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_OTHER)
			
			/*
			string.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_STRING)
			if ((i > 3) and (i < 7))
			{
				my_printf("%s\n", pgn_value);
				if (i == 4) strcpy(white_player, pgn_value);
				if (i == 5) strcpy(black_player, pgn_value);
				if (i == 6) strcpy(result, pgn_value);
			}
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_RBRACKET)
			i++;
			get_next_pgn_token(f);
		}
		ngame++;
		fen_to_board(start_fen, TRUE);
		ngame_moves = 0;
		gen_root();
		while(TRUE)
		{
			if (pgn_kind == PGN_RESULT) break;
			while(TRUE)
			{
				if (pgn_kind == PGN_OTHER) break;
				get_next_pgn_token(f);
			}
			base = capt_top;
			BUG((imove = search_san(pgn_value)) == INVALID)
			do_move(base[imove]);
			capt_top = base;

			BUG(ngame_moves >= NGAME_MOVES_MAX)
			game_moves[ngame_moves] = imove;
			ngame_moves++;
			get_next_pgn_token(f);
		}
		get_next_pgn_token(f);
		undo_some_moves(ngame_moves);
		bin_to_asc(ngame_moves, game_moves, game_string);
		r = crc((uchar_t *) game_string) % LONG_MAX;
		if (btree_search(r))
		{
			my_printf("This game is already in the database!\n");
		}
		else
		{
			FEOF(fbin)
			BUG((i = ftell(fbin)) < 0)
			btree_insert(r, i);
			FWRITE_STRING(fbin, white_player)
			FWRITE_STRING(fbin, black_player)
			FWRITE_STRING(fbin, result)
			FWRITE_STRING(fbin, game_string)
		}
	}
	close_dbase();
}

local void comp_key(FILE *);
local void comp_key(FILE *f)
{
	int imove;
	long i;
	int game_moves[NGAME_MOVES_MAX];
	char game_string[NGAME_MOVES_MAX];
	int ngame_moves;
	move_t *base;
	FILE *f_bin;
	
	BUG((f_bin = fopen("keys.bin", "w")) == NULL)
	get_next_pgn_token(f);
	while(TRUE)
	{
		/*
		tags.
		*/
		
		i = 0;
		if (pgn_kind == PGN_EOF) break;
		while(TRUE)
		{
			if (pgn_kind != PGN_LBRACKET) break;
			
			/*
			symbol.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_OTHER)
			
			/*
			string.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_STRING)
			if ((i > 3) and (i < 7))
			{
				my_printf("%s\n", pgn_value);
				fprintf(f_bin, "%s\n", pgn_value);
			}
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_RBRACKET)
			i++;
			get_next_pgn_token(f);
		}
		ngame++;
		fen_to_board(start_fen, TRUE);
		ngame_moves = 0;
		gen_root();
		while(TRUE)
		{
			if (pgn_kind == PGN_RESULT) break;
			while(TRUE)
			{
				if (pgn_kind == PGN_OTHER) break;
				get_next_pgn_token(f);
			}
			base = capt_top;
			if ((imove = search_san(pgn_value)) == INVALID)
			{
				my_printf("%s\n", pgn_value);
				FATAL_ERROR("invalid move?")
			}
			do_move(base[imove]);
			capt_top = base;

			BUG(ngame_moves >= NGAME_MOVES_MAX)
			game_moves[ngame_moves] = imove;
			ngame_moves++;
			get_next_pgn_token(f);
		}
		get_next_pgn_token(f);
		undo_some_moves(ngame_moves);
		bin_to_asc(ngame_moves, game_moves, game_string);
		fprintf(f_bin, "%s\n", game_string);
	}
	FCLOSE(f_bin)
}

#define HITS_MAX 256

local void anal_pgn(FILE *);
local void anal_pgn(FILE *f)
{
	int imove;
	int i;
	int game_moves[NGAME_MOVES_MAX], dbase_moves[NGAME_MOVES_MAX];
	char game_string[NGAME_MOVES_MAX], key_string[NGAME_MOVES_MAX], eco_string[NGAME_MOVES_MAX];
	int ngame_moves, ndbase_moves;
	move_t *base;
	FILE *f_bin;
	char white_player[PGN_LINE_MAX], black_player[PGN_LINE_MAX], result[PGN_LINE_MAX];
	hash_key_t game_key;
	int game_count[2];
	int nmoves;
	int ihit, nhits;
	move_t moves_hit[HITS_MAX];
	long won_hits[HITS_MAX];
	long draw_hits[HITS_MAX];
	long lost_hits[HITS_MAX];
	long ngame;
	clock_t t1, t2;
	int nlist;
	size_t len_game, len_key, len_eco;
	uchar_t magic, version;
	long feof;
	
	get_next_pgn_token(f);
	while(TRUE)
	{
		/*
		tags.
		*/
		
		i = 0;
		while(TRUE)
		{
			if (pgn_kind == PGN_EOF) return;
			if (pgn_kind != PGN_LBRACKET) break;
			
			/*
			symbol.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_OTHER)
			
			/*
			string.
			*/
			
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_STRING)
			if ((i > 3) and (i < 7))
			{
				my_printf("%s\n", pgn_value);
			}
			get_next_pgn_token(f);
			BUG(pgn_kind != PGN_RBRACKET)
			i++;
			get_next_pgn_token(f);
		}
		fen_to_board(start_fen, TRUE);
		ngame_moves = 0;
		gen_root();
		while(TRUE)
		{
			if (pgn_kind == PGN_RESULT) break;
			while(TRUE)
			{
				if (pgn_kind == PGN_OTHER) break;
				get_next_pgn_token(f);
			}
			display_board();
			my_printf("move played %s\n", pgn_value);
			
			/*
			search the current position in the pgn data base
			*/
			
			game_key = node->node_total_key;
			game_count[WHITE_COLOUR] = tot_count[WHITE_COLOUR];
			game_count[BLACK_COLOUR] = tot_count[BLACK_COLOUR];
			
			/*
			undo all moves for the starting position.
			*/
			
			undo_some_moves(ngame_moves);
			
			/*
			search the pgn data base
			*/
			
			nhits = 0;
			ngame = 0;
			nlist = 0;
			t1 = clock();
			BUG((f_bin = fopen("pgn.bin", "rb")) == NULL)
			FEOF(f_bin)
			BUG((feof = ftell(f_bin)) < 0)
			FBOF(f_bin)
			FREAD(f_bin, &magic, uchar_t, 1)
			FREAD(f_bin, &version, uchar_t, 1)
			my_printf("magic is %u version is %u\n", (uint_t) magic, (uint_t) version);
			while(TRUE)
			{
#if MACHINE == MAC
				if (Button()) break;
#endif
				if (ftell(f_bin) == feof) break;
				ngame++;
				FREAD_STRING(f_bin, white_player)
				FREAD_STRING(f_bin, black_player)
				FREAD_STRING(f_bin, result)
				FREAD_STRING(f_bin, game_string)
				asc_to_bin(&ndbase_moves, dbase_moves, game_string);
				
				for (i = 0; i < ndbase_moves; i++)
				{
					imove = dbase_moves[i];
					base = capt_top;
					nmoves = gen_all_moves();
					capt_top = base;
					if (imove >= nmoves)
					{
						display_board();
						if (my_colour == WHITE_COLOUR) my_printf("W\n"); else my_printf("B\n");
					}
					BUG(imove >= nmoves)
					if ((tot_count[WHITE_COLOUR] == game_count[WHITE_COLOUR]) and
					    (tot_count[BLACK_COLOUR] == game_count[BLACK_COLOUR]) and
					    EQUAL_KEY(node->node_total_key, game_key))
					{
						if (nlist < 10)
						{
							move_to_san_string(base[imove], FALSE);
							my_printf("also in %s- %s %s move %s\n",
								white_player, black_player, result, move_string);
							nlist++;
						}
						
						/*
						search
						*/
						
						for (ihit = 0; ihit < nhits; ihit++)
							if (moves_hit[ihit] == base[imove]) break;
						if (ihit >= nhits)
						{
							BUG(nhits >= HITS_MAX)
							moves_hit[nhits] = base[imove];
							won_hits[nhits] = draw_hits[nhits] = lost_hits[nhits] = 0;
							ihit = nhits;
							nhits++;
						}
						if (strncmp(result, "1-0", 3) == 0)
							won_hits[ihit]++;
						else if (strncmp(result, "1/2-1/2", 7) == 0)
							draw_hits[ihit]++;
						else if (strncmp(result, "0-1", 3) == 0)
							lost_hits[ihit]++;
						break;
					}
					
					/*
					short cut, can be done smarter..
					*/
					
					if (tot_count[WHITE_COLOUR] < game_count[WHITE_COLOUR]) break;
					if (tot_count[BLACK_COLOUR] < game_count[BLACK_COLOUR]) break;
					
					do_move(base[imove]);
				}
				undo_some_moves(i);
			}
			FCLOSE(f_bin)
			t2 = clock();
			if ((t2 - t1) > 0)
			{
				my_printf("games per sec=%.2f\n",
					(double) ngame / ((double) (t2 - t1) / CLOCKS_PER_SEC));
			}
			
			/*
			now redo all moves
			*/
			
			do_some_moves(ngame_moves, game_moves);
			
			/*
			search the ECO key. Match it to the longest string.
			*/
			
			strcpy(eco_string, "");
			len_eco = 0;
			bin_to_asc(ngame_moves, game_moves, game_string);
			len_game = strlen(game_string);
			
			BUG((f_bin = fopen("keys.bin", "r")) == NULL)
			while(TRUE)
			{
#if MACHINE == MAC
				if (Button()) break;
#endif
				if (my_fgets(white_player, PGN_LINE_MAX, f_bin) == NULL) break;
				BUG(my_fgets(black_player, PGN_LINE_MAX, f_bin) == NULL)
				BUG(my_fgets(result, PGN_LINE_MAX, f_bin) == NULL)
				BUG(my_fgets(key_string, NGAME_MOVES_MAX, f_bin) == NULL)
				len_key = strlen(key_string);
				
				if (len_game >= len_key)
				{
					if (strncmp(game_string, key_string, len_key) == 0)
					{
						/*
						possible hit.
						*/
						
						if (len_key > len_eco)
						{
							strcpy(eco_string, white_player);
							len_eco = len_key;
						}
					}
				}
			}
			FCLOSE(f_bin)
			
			/*
			display the summary
			*/
			
			my_printf("ECO key: %s\n", eco_string);
			if (nhits == 0)
				my_printf("no hits!\n");
			else
			{
				long total;
				my_printf("--move-- -games-- --won--- --draw-- --lost--\n");
				for (ihit = 0; ihit < nhits; ihit++)
				{
					move_to_san_string(moves_hit[ihit], TRUE);
					total = won_hits[ihit] + draw_hits[ihit] + lost_hits[ihit];
					my_printf("%8s %8ld %8.2f %8.2f %8.2f\n",
						move_string, total,
						(double) won_hits[ihit] / total * 100,
						(double) draw_hits[ihit] / total * 100,
						(double) lost_hits[ihit] / total * 100);
				}
			}
			
			base = capt_top;
			BUG((imove = search_san(pgn_value)) == INVALID)
			capt_top = base;
			do_move(base[imove]);
			
			BUG(ngame_moves >= NGAME_MOVES_MAX)
			game_moves[ngame_moves] = imove;
			ngame_moves++;
			get_next_pgn_token(f);
		}
		get_next_pgn_token(f);
		undo_some_moves(ngame_moves);
	}
}

void pgn(void)
{
	FILE *f;
	char line[LINE_MAX], comm[LINE_MAX], arg[LINE_MAX];
	int narg;
	
	while(TRUE)
	{
		my_printf("pgn> ");
		if (gets(line) == NULL) exit(EXIT_FAILURE);
		if ((narg = sscanf(line, "%s%s", comm, arg)) == 0) continue;
		if (strcmp(comm, "quit") == 0) break;
		if (strcmp(comm, "comp") == 0)
		{
			if (narg != 2)
			{
				my_printf("comp should be followed by a file name!\n");
				continue;
			}
			if ((f = fopen(arg, "r")) == NULL)
			{
				my_printf("cound not open %s!\n", arg);
				continue;
			}
			pbuf = buf;
			*buf = '\0';
			in_comment = FALSE;
			ngame = 0;
			comp_pgn(f);
			FCLOSE(f)
			my_printf("read %ld games\n", ngame);
			continue;
		}
		if (strcmp(comm, "keys") == 0)
		{
			if (narg != 2)
			{
				my_printf("key should be followed by a file name!\n");
				continue;
			}
			if ((f = fopen(arg, "r")) == NULL)
			{
				my_printf("could not open %s!\n", arg);
				continue;
			}
			pbuf = buf;
			*buf = '\0';
			in_comment = FALSE;
			ngame = 0;
			comp_key(f);
			FCLOSE(f)
			continue;
		}
		if (strcmp(comm, "anal") == 0)
		{
			if (narg != 2)
			{
				my_printf("anal should be followed by a file name!\n");
				continue;
			}
			if ((f = fopen(arg, "r")) == NULL)
			{
				my_printf("could not open %s!\n", arg);
				continue;
			}
			pbuf = buf;
			*buf = '\0';
			in_comment = FALSE;
			ngame = 0;
			anal_pgn(f);
			FCLOSE(f)
			continue;
		}
		my_printf("possible commands are 'comp', 'anal', 'keys' and 'quit'\n");
	}
}

char dbase_line[DBASE_LINE_MAX];
char *dbase_records[DBASE_LINE_MAX];
int ndbase_records;

/*
split_dbase_records splits a string into smaller strings.
*/

local void split_dbase_records(void);
local void split_dbase_records(void)
{
	char *string = dbase_line;
	
	ndbase_records = 0;
	while(*string != '\0')
	{
		dbase_records[ndbase_records] = string;
		while(*string++ != '\0');
		ndbase_records++;
	}
}

local int search_dbase_records(char *);
local int search_dbase_records(char *string)
{
	int irecord;
	
	for (irecord = 1; irecord < ndbase_records; irecord++)
	{
		if (strcmp(string, dbase_records[irecord]) == 0)
			return(irecord);
	}
	return(INVALID);
}

/*
join_dbase_records combines smaller strings to one string.
*/

local void join_dbase_records(void);
local void join_dbase_records(void)
{
	char *p;
	char *string = dbase_line;
	int irecord;
	
	for (irecord = 0; irecord < ndbase_records; irecord++)
	{
		p = dbase_records[irecord];
		while((*string++ = *p++) != '\0');
	}
	*string = '\0';
}

int search_dbase(void)
{
	long r;
	int found;
	char string[DBASE_LINE_MAX];

	found = FALSE;
	open_dbase("dbase", DBASE_MAGIC, DBASE_VERSION);
	board_to_fen(string, FALSE);
	r = crc((uchar_t *) string) % LONG_MAX;
	if (btree_search(r))
	{
		found = TRUE;
		FSEEK(fbin, btree_node.dbase_offset[btree_index])
		FREAD(fbin, dbase_line, char, DBASE_LINE_MAX)
		split_dbase_records();
	}
	close_dbase();
	return(found);
}

local void dump_dbase(long);
local void dump_dbase(long btree_temp_offset)
{
	int i;
	btree_node_t btree_temp;
	int irecord;
	
	if (btree_temp_offset == LONG_MAX) return;
	FSEEK(fkey, btree_temp_offset)
	FREAD(fkey, &btree_temp, btree_node_t, 1)
	for (i = 0; i < BTREE_ORDER + 1; i++)
	{
		if (btree_temp.btree_entry[i] != LONG_MAX)
		{
			FSEEK(fbin, btree_temp.dbase_offset[i])
			FREAD(fbin, dbase_line, char, DBASE_LINE_MAX)
			split_dbase_records();
			fen_to_board(dbase_records[0], FALSE);
			BUG((crc((uchar_t *) dbase_records[0]) % LONG_MAX) != btree_temp.btree_entry[i])
			display_board();
			for (irecord = 1; irecord < ndbase_records; irecord++)
				my_printf("irecord=%d %s\n", irecord, dbase_records[irecord]);
			dump_dbase(btree_temp.btree_child[i]);
		}
	}
}

void dbase(char *name)
{
	FILE *f;
	char fen_string[DBASE_LINE_MAX];
	char info[DBASE_LINE_MAX];
	long r, i;

#define DEBUG_DBASE
#ifdef DEBUG_DBASE
	remove("dbase.key");
	remove("dbase.bin");
#endif
	open_dbase("dbase", DBASE_MAGIC, DBASE_VERSION);
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(fen_string, DBASE_LINE_MAX, f) != NULL)
	{
		BUG(my_fgets(info, DBASE_LINE_MAX, f) == NULL)
		
		fen_to_board(fen_string, FALSE);
		if (options.print_dbase > 0)
		{
			display_board();
			my_printf("%s\n", info);
		}
		
		r = crc((uchar_t *) fen_string) % LONG_MAX;
		if (btree_search(r))
		{
			FSEEK(fbin, btree_node.dbase_offset[btree_index])
			FREAD(fbin, dbase_line, char, DBASE_LINE_MAX)
			split_dbase_records();
			if (strcmp(dbase_records[0], fen_string) != 0)
			{
				my_printf("Use 64 bit CRC! Complain to G. Wiesenekker\n");
				FATAL_ERROR("32 bit CRC problem")
			}
			if (search_dbase_records(info) == INVALID)
			{
				dbase_records[ndbase_records++] = info;
				join_dbase_records();
				FSEEK(fbin, btree_node.dbase_offset[btree_index])
				FWRITE(fbin, dbase_line, char, DBASE_LINE_MAX)
			}
		}
		else
		{
			FEOF(fbin)
			BUG((i = ftell(fbin)) < 0)
			btree_insert(r, i);
			dbase_records[0] = fen_string;
			dbase_records[1] = info;
			ndbase_records = 2;
			join_dbase_records();
			FWRITE(fbin, dbase_line, char, DBASE_LINE_MAX)
		}
	}
	FCLOSE(f)
	FBOF(fkey)
	FREAD(fkey, &btree_node_offset, long, 1)
	my_printf("dumping dbase\n");
	if (options.print_dbase > 0)
	{
		dump_dbase(btree_node_offset);
	}
	close_dbase();
}
